//-----------------------------------------------------------------------------
// 敪3֐
// Copyright (c) 2008  N099
//-----------------------------------------------------------------------------
#include <xmmintrin.h>
#include <stdlib.h>
#include <math.h>

#ifndef PIXEL_YC
typedef	struct {
    short y,cb,cr;
} PIXEL_YC;
#endif

#include "cubic.h"

//constant
static const __m128 xc0 = { 0.5f, 0.5f, 0.5f, 0.5f };
static const __m128 xc1 = { 1.0f, 1.0f, 1.0f, 1.0f };
static const __m128 xc2 = { 2.0f, 2.0f, 2.0f, 2.0f };

//internal
static void cb_interporate_h_yc(PIXEL_YC *yc, const __m128 *xsrc,
                                const p_weight *we, int w, int rw);
static void cb_decimate_h_yc(PIXEL_YC *yc, const __m128 *xsrc,
                             const p_weight *we, int w, int rw);
static void cb_interporate_v_yc(__m64 *myc, const __m128 *xsrc,
                                const p_weight *we, int h, int rh, int ws);
static void cb_decimate_v_yc(__m64 *myc, const __m128 *xsrc,
                             const p_weight *we, int h, int rh, int ws);

//dݐݒ
void cb_set_weight(p_weight *we, int type, int b, int c)
{
    float fb = (float)(b)/128.0f;
    float fc = (float)(c)/128.0f;

    //Mitchell-Netravali filter
    // Mitchell, D. P. and Netravali, A. N. 1988.
    // Reconstruction filters in computer-graphics.
    // SIGGRAPH Comput. Graph. 22, 4 (Aug. 1988), 221-228.
    we->a = _mm_set1_ps(( 12.0f -  9.0f*fb - 6.0f*fc)/6.0f);
    we->b = _mm_set1_ps((-18.0f + 12.0f*fb + 6.0f*fc)/6.0f);
    we->c = _mm_set1_ps(   0.0f);
    we->d = _mm_set1_ps((  6.0f -  2.0f*fb          )/6.0f);
    
    we->e = _mm_set1_ps((-      fb -  6.0f*fc)/6.0f);
    we->f = _mm_set1_ps((  6.0f*fb + 30.0f*fc)/6.0f);
    we->g = _mm_set1_ps((-12.0f*fb - 48.0f*fc)/6.0f);
    we->h = _mm_set1_ps((  8.0f*fb + 24.0f*fc)/6.0f);
    
    _mm_empty();
}

//PIXEL_YC->floatϊij
void cb_cnv_h_yc(const PIXEL_YC *yc, __m128 *xsrc, int w)
{
    int i;
    __m128 xyc0;

    /* [4sNZ͗אڃf[^Rs[Ă(w = w+4) */
    xyc0 = _mm_cvtpi16_ps(*(const __m64*)yc);
    _mm_store_ps((float*)xsrc++, xyc0);
    _mm_store_ps((float*)xsrc++, xyc0);

    for(i=0; i<w; ++i){
        xyc0 = _mm_cvtpi16_ps(*(const __m64*)(yc++));
        _mm_store_ps((float*)xsrc++, xyc0);
    }

    //E[
    _mm_store_ps((float*)xsrc++, xyc0);
    _mm_store_ps((float*)xsrc++, xyc0);

    _mm_empty();
 }

//TCYPIXEL_YCɕϊij
void cb_hresize_yc(PIXEL_YC *yc, const __m128 *xsrc, const p_weight *we,
                          int w, int rw)
{
    if(w > rw){
        cb_decimate_h_yc(yc, xsrc, we, w, rw);
    }else{
        cb_interporate_h_yc(yc, xsrc, we, w, rw);
    }
}

//敪3֐ɂԁij
void cb_interporate_h_yc(PIXEL_YC *yc, const __m128 *xsrc, const p_weight *we,
                         int w, int rw)
{
    int i,p;
    __m128 xs0,xs1,xs2,xs3,xa1,xb1,xa2,xb2,xa3,xb3,xss,xdp;
    
    //sb`̌vZ
    xdp = _mm_div_ss( _mm_cvtsi32_ss(_mm_setzero_ps(), w),
                      _mm_cvtsi32_ss(_mm_setzero_ps(), rw) );
    
    for(i=0; i<rw; ++i){
        xa3 = _mm_cvtsi32_ss(xc0, i);
        xa3 = _mm_add_ss(xa3, xc0);

        xa1 = _mm_mul_ss(xdp, xa3);
        xa1 = _mm_add_ss(xa1, xc0);
        p = _mm_cvttss_si32(xa1);

        xs0 = _mm_load_ps((const float*)(xsrc + p));
        xs1 = _mm_load_ps((const float*)(xsrc + p + 1));
        xs2 = _mm_load_ps((const float*)(xsrc + p + 2));
        xs3 = _mm_load_ps((const float*)(xsrc + p + 3));
        
        xa3 = _mm_cvtsi32_ss(xa3, p);
        xb1 = _mm_sub_ss(xc1, xa1);
        xa1 = _mm_sub_ss(xa1, xa3);
        xb1 = _mm_add_ss(xb1, xa3);

        xb1 = _mm_shuffle_ps(xb1, xb1, 0);
        xa1 = _mm_shuffle_ps(xa1, xa1, 0);

        xb3 = _mm_add_ps(we->d, _mm_mul_ps(we->c, xb1));
        xb2 = _mm_mul_ps(xb1, xb1);
        xb3 = _mm_add_ps(xb3, _mm_mul_ps(we->b, xb2));
        xb2 = _mm_mul_ps(xb2, xb1);
        xb3 = _mm_add_ps(xb3, _mm_mul_ps(we->a, xb2));
        xss = _mm_mul_ps(xb3, xs2);
        
        xa3 = _mm_add_ps(we->d, _mm_mul_ps(we->c, xa1));
        xa2 = _mm_mul_ps(xa1, xa1);
        xa3 = _mm_add_ps(xa3, _mm_mul_ps(we->b, xa2));
        xa2 = _mm_mul_ps(xa2, xa1);
        xa3 = _mm_add_ps(xa3, _mm_mul_ps(we->a, xa2));
        xa3 = _mm_mul_ps(xa3, xs1);
        xss = _mm_add_ps(xss, xa3);
        
        xb1 = _mm_add_ps(xb1, xc1);
        xb3 = _mm_add_ps(we->h, _mm_mul_ps(we->g, xb1));
        xb2 = _mm_mul_ps(xb1, xb1);
        xb3 = _mm_add_ps(xb3, _mm_mul_ps(we->f, xb2));
        xb2 = _mm_mul_ps(xb2, xb1);
        xb3 = _mm_add_ps(xb3, _mm_mul_ps(we->e, xb2));
        xb3 = _mm_mul_ps(xb3, xs3);
        xss = _mm_add_ps(xss, xb3);

        xa1 = _mm_add_ps(xa1, xc1);
        xa3 = _mm_add_ps(we->h, _mm_mul_ps(we->g, xa1));
        xa2 = _mm_mul_ps(xa1, xa1);
        xa3 = _mm_add_ps(xa3, _mm_mul_ps(we->f, xa2));
        xa2 = _mm_mul_ps(xa2, xa1);
        xa3 = _mm_add_ps(xa3, _mm_mul_ps(we->e, xa2));
        xa3 = _mm_mul_ps(xa3, xs0);
        xss = _mm_add_ps(xss, xa3);
        
        *(__m64*)yc++ = _mm_cvtps_pi16(xss);
    }
    
    _mm_empty();
}

//敪3֐ɂkij
void cb_decimate_h_yc(PIXEL_YC *yc, const __m128 *xsrc, const p_weight *we,
                      int w, int rw)
{
    int i,a,b,aa,bb;
    __m128 xs0,xs1,xa1,xb1,xa2,xb2,xa3,xb3,xpf,xss,xrp;
    
    //ItZbg1pixeli߂
    xsrc += 1;
    
    //sb`̌vZitj
    xrp = _mm_div_ss( _mm_cvtsi32_ss(_mm_setzero_ps(), rw),
                      _mm_cvtsi32_ss(_mm_setzero_ps(), w) );
    xrp = _mm_shuffle_ps(xrp, xrp, 0);
    
    for(i=0; i<rw; ++i){
        xpf = _mm_cvtsi32_ss(xc0, i);
        xpf = _mm_add_ss(xpf, xc0);

        xpf = _mm_mul_ss(xpf, _mm_rcp_ss(xrp));
        xpf = _mm_add_ss(xpf, xc0);
        
        a = _mm_cvttss_si32(xpf);
        b = a + 1;
        
        xa1 = _mm_cvtsi32_ss(xa1, a);
        xb1 = _mm_cvtsi32_ss(xb1, b);
        xa1 = _mm_mul_ss(_mm_sub_ss(xpf, xa1), xrp);
        xb1 = _mm_mul_ss(_mm_sub_ss(xb1, xpf), xrp);
        
        xss = _mm_setzero_ps();
        
        while(_mm_ucomile_ss(xa1, xc1)){
            aa  = a > 0 ? a : -a;
            xs0 = _mm_load_ps((const float*)(xsrc + aa));
            
            xa1 = _mm_shuffle_ps(xa1, xa1, 0);
        
            xa3 = _mm_add_ps(we->d, _mm_mul_ps(we->c, xa1));
            xa2 = _mm_mul_ps(xa1, xa1);
            xa3 = _mm_add_ps(xa3, _mm_mul_ps(we->b, xa2));
            xa2 = _mm_mul_ps(xa2, xa1);
            xa3 = _mm_add_ps(xa3, _mm_mul_ps(we->a, xa2));
            xa3 = _mm_mul_ps(xa3, xs0);
            xss = _mm_add_ps(xss, xa3);

            --a;
            xa1 = _mm_cvtsi32_ss(xa1, a);
            xa1 = _mm_mul_ss(_mm_sub_ss(xpf, xa1), xrp);
        }
        while(_mm_ucomile_ss(xa1, xc2)){
            aa  = a > 0 ? a : -a;
            xs0 = _mm_load_ps((const float*)(xsrc + aa));

            xa1 = _mm_shuffle_ps(xa1, xa1, 0);
            
            xa3 = _mm_add_ps(we->h, _mm_mul_ps(we->g, xa1));
            xa2 = _mm_mul_ps(xa1, xa1);
            xa3 = _mm_add_ps(xa3, _mm_mul_ps(we->f, xa2));
            xa2 = _mm_mul_ps(xa2, xa1);
            xa3 = _mm_add_ps(xa3, _mm_mul_ps(we->e, xa2));
            xa3 = _mm_mul_ps(xa3, xs0);
            xss = _mm_add_ps(xss, xa3);
            
            --a;
            xa1 = _mm_cvtsi32_ss(xa1, a);
            xa1 = _mm_mul_ss(_mm_sub_ss(xpf, xa1), xrp);
        }
        while(_mm_ucomile_ss(xb1, xc1)){
            bb  = b < w ? b : 2*w - b;
            xs1 = _mm_load_ps((const float*)(xsrc + bb));

            xb1 = _mm_shuffle_ps(xb1, xb1, 0);
            
            xb3 = _mm_add_ps(we->d, _mm_mul_ps(we->c, xb1));
            xb2 = _mm_mul_ps(xb1, xb1);
            xb3 = _mm_add_ps(xb3, _mm_mul_ps(we->b, xb2));
            xb2 = _mm_mul_ps(xb2, xb1);
            xb3 = _mm_add_ps(xb3, _mm_mul_ps(we->a, xb2));
            xb3 = _mm_mul_ps(xb3, xs1);
            xss = _mm_add_ps(xss, xb3);

            ++b;
            xb1 = _mm_cvtsi32_ss(xb1, b);
            xb1 = _mm_mul_ss(_mm_sub_ss(xb1, xpf), xrp);
        }
        while(_mm_ucomile_ss(xb1, xc2)){
            bb  = b < w ? b : 2*w - b;
            xs1 = _mm_load_ps((const float*)(xsrc + bb));

            xb1 = _mm_shuffle_ps(xb1, xb1, 0);
            
            xb3 = _mm_add_ps(we->h, _mm_mul_ps(we->g, xb1));
            xb2 = _mm_mul_ps(xb1, xb1);
            xb3 = _mm_add_ps(xb3, _mm_mul_ps(we->f, xb2));
            xb2 = _mm_mul_ps(xb2, xb1);
            xb3 = _mm_add_ps(xb3, _mm_mul_ps(we->e, xb2));
            xb3 = _mm_mul_ps(xb3, xs1);
            xss = _mm_add_ps(xss, xb3);
            
            ++b;
            xb1 = _mm_cvtsi32_ss(xb1, b);
            xb1 = _mm_mul_ss(_mm_sub_ss(xb1, xpf), xrp);
        }
        
        *(__m64*)yc++ = _mm_cvtps_pi16(_mm_mul_ps(xss, xrp));
    }
    
    _mm_empty();
}


//PIXEL_YC->floatϊij
void cb_cnv_v_yc(const __m64 *myc, __m128 *xsrc, int h, int ws)
{
    int i;
    __m128 xyc0;

    /* ㉺[2*2sNZ͗אڃf[^Rs[Ă(h = h+4) */
    xyc0 = _mm_cvtpi16_ps(*myc);
    _mm_store_ps((float*)xsrc++, xyc0);
    _mm_store_ps((float*)xsrc++, xyc0);

    //
    for(i=0; i<h; ++i){
        xyc0 = _mm_cvtpi16_ps(*myc);
        _mm_store_ps((float*)xsrc++, xyc0);
        myc += ws;
    }

    //[
    _mm_store_ps((float*)xsrc++, xyc0);
    _mm_store_ps((float*)xsrc++, xyc0);
    
    _mm_empty();
}

//TCYPIXEL_YCɕϊij
void cb_vresize_yc(__m64 *myc, const __m128 *xsrc, const p_weight *we,
                          int h, int rh, int ws)
{
    if(h > rh){
        cb_decimate_v_yc(myc, xsrc, we, h, rh, ws);
    }else{
        cb_interporate_v_yc(myc, xsrc, we, h, rh, ws);
    }
}

//敪3֐ɂԁij
void cb_interporate_v_yc(__m64 *myc, const __m128 *xsrc,
                         const p_weight *we, int h, int rh, int ws)
{
    int i,p;
    __m128 xs0,xs1,xs2,xs3,xa1,xb1,xa2,xb2,xa3,xb3,xss,xdp;
    
    //sb`̌vZ
    xdp = _mm_div_ss( _mm_cvtsi32_ss(_mm_setzero_ps(), h),
                      _mm_cvtsi32_ss(_mm_setzero_ps(), rh) );
    
    for(i=0; i<rh; ++i){
        xa3 = _mm_cvtsi32_ss(xdp, i);
        xa3 = _mm_add_ss(xa3, xc0);

        xa1 = _mm_mul_ss(xdp, xa3);
        xa1 = _mm_add_ss(xa1, xc0);
        xb1 = xc1;
        p = _mm_cvttss_si32(xa1);

        xs0 = _mm_load_ps((const float*)(xsrc + p));
        xs1 = _mm_load_ps((const float*)(xsrc + p + 1));
        xs2 = _mm_load_ps((const float*)(xsrc + p + 2));
        xs3 = _mm_load_ps((const float*)(xsrc + p + 3));
        
        xa3 = _mm_cvtsi32_ss(xa3, p);
        xb1 = _mm_sub_ss(xb1, xa1);
        xa1 = _mm_sub_ss(xa1, xa3);
        xb1 = _mm_add_ss(xb1, xa3);

        xb1 = _mm_shuffle_ps(xb1, xb1, 0);
        xa1 = _mm_shuffle_ps(xa1, xa1, 0);

        xb3 = _mm_add_ps(we->d, _mm_mul_ps(we->c, xb1));
        xb2 = _mm_mul_ps(xb1, xb1);
        xb3 = _mm_add_ps(xb3, _mm_mul_ps(we->b, xb2));
        xb2 = _mm_mul_ps(xb2, xb1);
        xb3 = _mm_add_ps(xb3, _mm_mul_ps(we->a, xb2));
        xss = _mm_mul_ps(xb3, xs2);
        
        xa3 = _mm_add_ps(we->d, _mm_mul_ps(we->c, xa1));
        xa2 = _mm_mul_ps(xa1, xa1);
        xa3 = _mm_add_ps(xa3, _mm_mul_ps(we->b, xa2));
        xa2 = _mm_mul_ps(xa2, xa1);
        xa3 = _mm_add_ps(xa3, _mm_mul_ps(we->a, xa2));
        xa3 = _mm_mul_ps(xa3, xs1);
        xss = _mm_add_ps(xss, xa3);
        
        xb1 = _mm_add_ps(xb1, xc1);
        xb3 = _mm_add_ps(we->h, _mm_mul_ps(we->g, xb1));
        xb2 = _mm_mul_ps(xb1, xb1);
        xb3 = _mm_add_ps(xb3, _mm_mul_ps(we->f, xb2));
        xb2 = _mm_mul_ps(xb2, xb1);
        xb3 = _mm_add_ps(xb3, _mm_mul_ps(we->e, xb2));
        xb3 = _mm_mul_ps(xb3, xs3);
        xss = _mm_add_ps(xss, xb3);

        xa1 = _mm_add_ps(xa1, xc1);
        xa3 = _mm_add_ps(we->h, _mm_mul_ps(we->g, xa1));
        xa2 = _mm_mul_ps(xa1, xa1);
        xa3 = _mm_add_ps(xa3, _mm_mul_ps(we->f, xa2));
        xa2 = _mm_mul_ps(xa2, xa1);
        xa3 = _mm_add_ps(xa3, _mm_mul_ps(we->e, xa2));
        xa3 = _mm_mul_ps(xa3, xs0);
        xss = _mm_add_ps(xss, xa3);
        
        *myc = _mm_cvtps_pi16(xss);
        myc += ws;
    }
    
    _mm_empty();
}

//敪3֐ɂkij
void cb_decimate_v_yc(__m64 *myc, const __m128 *xsrc,
                      const p_weight *we, int h, int rh, int ws)
{
    int i,a,b,aa,bb;
    __m128 xs0,xs1,xa1,xb1,xa2,xb2,xa3,xb3,xpf,xss,xrp;
    
    //ItZbg1si߂
    xsrc += 1;
    
    //sb`̌vZitj
    xrp = _mm_div_ss( _mm_cvtsi32_ss(_mm_setzero_ps(), rh),
                      _mm_cvtsi32_ss(_mm_setzero_ps(), h) );
    xrp = _mm_shuffle_ps(xrp, xrp, 0);
    
    for(i=0; i<rh; ++i){
        xpf = _mm_cvtsi32_ss(xc0, i);
        xpf = _mm_add_ss(xpf, xc0);

        xpf = _mm_mul_ss(xpf, _mm_rcp_ss(xrp));
        xpf = _mm_add_ss(xpf, xc0);
        
        a = _mm_cvttss_si32(xpf);
        b = a + 1;
        
        xa1 = _mm_cvtsi32_ss(xa1, a);
        xb1 = _mm_cvtsi32_ss(xb1, b);
        xa1 = _mm_mul_ss(_mm_sub_ss(xpf, xa1), xrp);
        xb1 = _mm_mul_ss(_mm_sub_ss(xb1, xpf), xrp);
        
        xss = _mm_setzero_ps();
        
        while(_mm_ucomile_ss(xa1, xc1)){
            aa  = a > 0 ? a : -a;
            xs0 = _mm_load_ps((const float*)(xsrc + aa));
            
            xa1 = _mm_shuffle_ps(xa1, xa1, 0);
        
            xa3 = _mm_add_ps(we->d, _mm_mul_ps(we->c, xa1));
            xa2 = _mm_mul_ps(xa1, xa1);
            xa3 = _mm_add_ps(xa3, _mm_mul_ps(we->b, xa2));
            xa2 = _mm_mul_ps(xa2, xa1);
            xa3 = _mm_add_ps(xa3, _mm_mul_ps(we->a, xa2));
            xa3 = _mm_mul_ps(xa3, xs0);
            xss = _mm_add_ps(xss, xa3);

            --a;
            xa1 = _mm_cvtsi32_ss(xa1, a);
            xa1 = _mm_mul_ss(_mm_sub_ss(xpf, xa1), xrp);
        }
        while(_mm_ucomile_ss(xa1, xc2)){
            aa  = a > 0 ? a : -a;
            xs0 = _mm_load_ps((const float*)(xsrc + aa));

            xa1 = _mm_shuffle_ps(xa1, xa1, 0);
            
            xa3 = _mm_add_ps(we->h, _mm_mul_ps(we->g, xa1));
            xa2 = _mm_mul_ps(xa1, xa1);
            xa3 = _mm_add_ps(xa3, _mm_mul_ps(we->f, xa2));
            xa2 = _mm_mul_ps(xa2, xa1);
            xa3 = _mm_add_ps(xa3, _mm_mul_ps(we->e, xa2));
            xa3 = _mm_mul_ps(xa3, xs0);
            xss = _mm_add_ps(xss, xa3);
            
            --a;
            xa1 = _mm_cvtsi32_ss(xa1, a);
            xa1 = _mm_mul_ss(_mm_sub_ss(xpf, xa1), xrp);
        }
        while(_mm_ucomile_ss(xb1, xc1)){
            bb  = b < h ? b : 2*h - b;
            xs1 = _mm_load_ps((const float*)(xsrc + bb));

            xb1 = _mm_shuffle_ps(xb1, xb1, 0);
            
            xb3 = _mm_add_ps(we->d, _mm_mul_ps(we->c, xb1));
            xb2 = _mm_mul_ps(xb1, xb1);
            xb3 = _mm_add_ps(xb3, _mm_mul_ps(we->b, xb2));
            xb2 = _mm_mul_ps(xb2, xb1);
            xb3 = _mm_add_ps(xb3, _mm_mul_ps(we->a, xb2));
            xb3 = _mm_mul_ps(xb3, xs1);
            xss = _mm_add_ps(xss, xb3);

            ++b;
            xb1 = _mm_cvtsi32_ss(xb1, b);
            xb1 = _mm_mul_ss(_mm_sub_ss(xb1, xpf), xrp);
        }
        while(_mm_ucomile_ss(xb1, xc2)){
            bb  = b < h ? b : 2*h - b;
            xs1 = _mm_load_ps((const float*)(xsrc + bb));

            xb1 = _mm_shuffle_ps(xb1, xb1, 0);
            
            xb3 = _mm_add_ps(we->h, _mm_mul_ps(we->g, xb1));
            xb2 = _mm_mul_ps(xb1, xb1);
            xb3 = _mm_add_ps(xb3, _mm_mul_ps(we->f, xb2));
            xb2 = _mm_mul_ps(xb2, xb1);
            xb3 = _mm_add_ps(xb3, _mm_mul_ps(we->e, xb2));
            xb3 = _mm_mul_ps(xb3, xs1);
            xss = _mm_add_ps(xss, xb3);
            
            ++b;
            xb1 = _mm_cvtsi32_ss(xb1, b);
            xb1 = _mm_mul_ss(_mm_sub_ss(xb1, xpf), xrp);
        }
        
        *myc = _mm_cvtps_pi16(_mm_mul_ps(xss, xrp));
        myc += ws;
    }
    
    _mm_empty();
}

